
# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-09-10T07:14:13-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "c12e1d6f1c13f5c94a0bbe953e9a7f1858f45c5e",
 "version": "0.1.0a32"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import sys

import dask.array
import dask.dataframe
import numpy
import pandas
import xarray

from tiled.readers.array import ArrayAdapter
from tiled.readers.dataframe import DataFrameAdapter
from tiled.readers.xarray import DataArrayAdapter, DatasetAdapter, VariableAdapter
from tiled.trees.in_memory import Tree


print("Generating large example data...", file=sys.stderr)
data = {
    "big_image": numpy.random.random((10_000, 10_000)),
    "small_image": numpy.random.random((100, 100)),
    "medium_image": numpy.random.random((1000, 1000)),
    "tiny_image": numpy.random.random((10, 10)),
    "tiny_cube": numpy.random.random((10, 10, 10)),
    "tiny_hypercube": numpy.random.random((10, 10, 10, 10, 10)),
    "high_entropy": numpy.random.random((100, 100)),
    "low_entropy": numpy.ones((100, 100)),
    "short_column": numpy.random.random(100),
    "long_column": numpy.random.random(100_000),
}
print("Done generating example data.", file=sys.stderr)

tree = Tree(
    {
        "big_image": ArrayAdapter.from_array(data["big_image"]),
        "small_image": ArrayAdapter.from_array(data["small_image"]),
        "tiny_image": ArrayAdapter.from_array(data["tiny_image"]),
        "tiny_cube": ArrayAdapter.from_array(data["tiny_cube"]),
        "tiny_hypercube": ArrayAdapter.from_array(data["tiny_hypercube"]),
        "low_entropy": ArrayAdapter.from_array(data["low_entropy"]),
        "high_entropy": ArrayAdapter.from_array(data["high_entropy"]),
        "short_table": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "A": data["short_column"],
                    "B": 2 * data["short_column"],
                    "C": 3 * data["short_column"],
                },
                index=pandas.Index(
                    numpy.arange(len(data["short_column"])), name="index"
                ),
            ),
            npartitions=1,
            metadata={"animal": "dog", "color": "red"},
        ),
        "long_table": DataFrameAdapter.from_pandas(
            pandas.DataFrame(
                {
                    "A": data["long_column"],
                    "B": 2 * data["long_column"],
                    "C": 3 * data["long_column"],
                },
                index=pandas.Index(
                    numpy.arange(len(data["long_column"])), name="index"
                ),
            ),
            npartitions=5,
            metadata={"animal": "dog", "color": "green"},
        ),
        "labeled_data": Tree(
            {
                "image_with_dims": VariableAdapter(
                    xarray.Variable(
                        data=dask.array.from_array(data["medium_image"]),
                        dims=["x", "y"],
                        attrs={"thing": "stuff"},
                    ),
                ),
            }
        ),
        "structured_data": Tree(
            {
                "image_with_coords": DataArrayAdapter(
                    xarray.DataArray(
                        xarray.Variable(
                            data=dask.array.from_array(data["medium_image"]),
                            dims=["x", "y"],
                            attrs={"thing": "stuff"},
                        ),
                        coords={
                            "x": dask.array.arange(len(data["medium_image"])) / 10,
                            "y": dask.array.arange(len(data["medium_image"])) / 50,
                        },
                    ),
                ),
                "xarray_dataset": DatasetAdapter(
                    xarray.Dataset(
                        {
                            "image": xarray.DataArray(
                                xarray.Variable(
                                    data=dask.array.from_array(data["medium_image"]),
                                    dims=["x", "y"],
                                    attrs={"thing": "stuff"},
                                ),
                                coords={
                                    "x": dask.array.arange(len(data["medium_image"]))
                                    / 10,
                                    "y": dask.array.arange(len(data["medium_image"]))
                                    / 50,
                                },
                            ),
                            "z": xarray.DataArray(
                                data=dask.array.ones((len(data["medium_image"]),))
                            ),
                        }
                    )
                ),
            },
            metadata={"animal": "cat", "color": "green"},
        ),
    },
)



import dask.array
import numpy
import pytest
import xarray.testing
import xarray

from ..readers.xarray import DataArrayAdapter, DatasetAdapter, VariableAdapter
from ..client import from_tree
from ..trees.in_memory import Tree
from ..client import xarray as xarray_client


array = numpy.random.random((10, 10))
tree = Tree(
    {
        "variable": VariableAdapter(
            xarray.Variable(
                data=dask.array.from_array(array),
                dims=["x", "y"],
                attrs={"thing": "stuff"},
            ),
        ),
        "data_array": DataArrayAdapter(
            xarray.DataArray(
                xarray.Variable(
                    data=dask.array.from_array(array),
                    dims=["x", "y"],
                    attrs={"thing": "stuff"},
                ),
                coords={
                    "x": dask.array.arange(len(array)),
                    "y": 10 * dask.array.arange(len(array)),
                },
            ),
        ),
        "dataset": DatasetAdapter(
            xarray.Dataset(
                {
                    "image": xarray.DataArray(
                        xarray.Variable(
                            data=dask.array.from_array(array),
                            dims=["x", "y"],
                            attrs={"thing": "stuff"},
                        ),
                        coords={
                            "x": dask.array.arange(len(array)),
                            "y": 10 * dask.array.arange(len(array)),
                        },
                    ),
                    "z": xarray.DataArray(data=dask.array.ones((len(array),))),
                }
            )
        ),
        "wide": DatasetAdapter(
            xarray.Dataset(
                {
                    f"column_{i:03}": xarray.DataArray(i * numpy.ones(10))
                    for i in range(500)
                }
            )
        ),
    }
)


@pytest.mark.parametrize("key", list(tree))
def test_xarrays(key):
    client = from_tree(tree)
    expected = tree[key].read()
    actual = client[key].read()
    xarray.testing.assert_equal(actual, expected)


def test_dataset_column_access():
    client = from_tree(tree)
    expected_dataset = tree["dataset"].read()
    client_dataset = client["dataset"]
    for col in expected_dataset:
        actual = client_dataset[col]
        expected = expected_dataset[col]
        xarray.testing.assert_equal(actual, expected)


def test_url_limit_handling():
    "Check that requests and split up to stay below the URL length limit."
    expected = tree["wide"].read()
    client = from_tree(tree)
    client["wide"].read()  # Dry run to run any one-off state-initializing requests.
    # Accumulate Requests here for later inspection.
    requests = []

    async def accumulate(request):
        # httpx.AsyncClient requires event hooks to be async functions.
        requests.append(request)

    client.context.event_hooks["request"].append(accumulate)
    actual = client["wide"].read()
    xarray.testing.assert_equal(actual, expected)
    normal_request_count = len(requests)
    original = xarray_client.URL_CHARACTER_LIMIT
    try:
        # It should never be necessary to tune this for real-world use, but we
        # use this knob as a way to test its operation.
        xarray_client.URL_CHARACTER_LIMIT = 200
        # The client will need to split this across more requests in order to
        # stay within the tighter limit.
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = client["wide"].read()
        xarray.testing.assert_equal(actual, expected)
        higher_request_count = len(requests)
        # Tighten even more.
        xarray_client.URL_CHARACTER_LIMIT = 100
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = client["wide"].read()
        xarray.testing.assert_equal(actual, expected)
        highest_request_count = len(requests)
    finally:
        # Restore default.
        xarray_client.URL_CHARACTER_LIMIT = original
    # The goal here is to test the *trend* not the specific values because the
    # number of requests may evolve as the library changes, but the trend should
    # hold.
    assert highest_request_count > higher_request_count > normal_request_count



from datetime import datetime

import numpy

from ..readers.array import ArrayAdapter
from ..trees.in_memory import Tree
from ..client import from_config
from ..server.authentication import create_refresh_token

arr = ArrayAdapter.from_array(numpy.ones((5, 5)))
tree_a = Tree({"A1": arr, "A2": arr})
tree_b = Tree({"B1": arr, "B2": arr})


def test_top_level_access_control():
    SECRET_KEY = "secret"
    config = {
        "authentication": {
            "secret_keys": [SECRET_KEY],
            "authenticator": "tiled.authenticators:DictionaryAuthenticator",
            "args": {"users_to_passwords": {"alice": "secret1", "bob": "secret2"}},
        },
        "access_control": {
            "access_policy": "tiled.trees.in_memory:SimpleAccessPolicy",
            "args": {"access_lists": {"alice": ["a"]}},
        },
        "trees": [
            {
                "tree": f"{__name__}:tree_a",
                "path": "/a",
            },
            {
                "tree": f"{__name__}:tree_b",
                "path": "/b",
            },
        ],
    }
    # Directly generate a refresh token.
    refresh_token = create_refresh_token(
        data={"sub": "alice"},
        session_id=0,
        session_creation_time=datetime.now(),
        secret_key=SECRET_KEY,
    )
    # Provide the refresh token in a token cache. The client
    # will use this to "refresh" and obtain an access token.
    client = from_config(
        config, username="alice", token_cache={"refresh_token": refresh_token}
    )
    assert "a" in client
    # Try accessing an item nested inside an item we can access.
    client["a"]["A1"]
    assert "b" not in client
